
"""
Compute CO2/NO2 mass along line or within area.
"""

import warnings

from scipy.constants import N_A

from shapely.geometry import LineString, Polygon

import numpy as np
import scipy.optimize
import xarray

import doas
import amrs.vis

import emiss.vis
import smartcarb.map

from . import misc
from . import functions


warnings.filterwarnings('ignore', 'Mean of empty slice')



def expand_dimension(data, dim, size):
        
    coords = {}
    for key in data.dims:
        if key == dim:
            coords[key] = np.arange(size)
        else:
            coords[key] = data[key]
    
    new = xarray.Dataset(coords=coords, attrs=data.attrs)
    
    for key in data:
        var = data[key]
        
        if dim in var.dims:
            fill_value = False if var.dtype == np.bool else np.nan
            values = np.concatenate([var.values, np.full(size-var[dim].size, fill_value, dtype=var.dtype)])     
        else:
            values = data[key].values
            
        new[key] = xarray.DataArray(values, dims=var.dims, attrs=var.attrs) 

    
    return new


def concat_polygons(polygons):
    size = max(p['pixels'].size for p in polygons)
    values = [expand_dimension(p, 'pixels', size) for p in polygons]
    return xarray.concat(values, dim='along')



def compute_mass(data, curve, field):
    """
    """
    distances = np.arange(0, np.nanmax(data.xp), 2e3)

    mass, mass_std = integrate_fields(data, field, curve, distances,
                                      tmax=100e3)

    valids = (distances > dmin) & np.isfinite(fluxes) & np.isfinite(fluxes_std)
    valids &= (distances <= dmax)

    mask = ~valids

    return np.ma.array(mass, mask=mask), np.ma.array(mass_std, mask=mask)



def integrate_fields(data, tracer, curve, distances, tmax):

    mass = []
    mass_std = []

    for d in distances:
        f,u = integrate_field(data, tracer, curve, d, tmax=tmax, show=False)
        mass.append(f)
        mass_std.append(u)

    return np.array(fluxes), np.array(fluxes_std)



def integrate_along_line(data, tracer, distances, curve=None, show=False,
                         method='sub-areas'):
    """
    distance from source origin
    """
    masses_true = []
    masses = []
    masses_std = []

    wind_eff = []
    wind_true = []

    for i in range(distances.size-1):

        results = extract_line_density(data, tracer, distances[i],
                                       distances[i+1],
                                       ya=None, yb=None, dy=5e3,
                                       method=method, show=False)

        if method == 'two-gauss':
            masses.append(results['co2-gauss'])
            masses_std.append(results['co2-gauss_std'])
        else:
            masses.append(results[method])
            masses_std.append(results['%s_std' % method])

        masses_true.append(results['density_true'])

        wind_eff.append(results['wind'])
        wind_true.append(results['wind_true'])

    wind_true = np.array(wind_true)
    wind_eff = np.array(wind_eff)

    return np.array(masses), np.array(masses_std), np.array(masses_true), \
           wind_eff, wind_true




def add_area(ax, data, curve, xa, xb, ya, yb, color='yellow',
             add_label=True, extra_width=5e3, lw=None,
             do_middle_lines=True, lds=None):
    
    # polygon (outside) 
    x_left = []
    x_right = []
    y_left = []
    y_right = []
    

    for i in range(xa.size):

        # interpolate distance from arc(t)
        xmin = curve.arc2parameter(xa[i])
        xmax = curve.arc2parameter(xb[i])
        y = np.array([ya[i], yb[i]])
            

        # compute normals
        (x0, x1), (y0, y1) = curve.compute_normal(xmin, t=y)
        (x3, x2), (y3, y2) = curve.compute_normal(xmax, t=y)

        if i == 0:
            x_right.extend([x0, x1, x2])
            x_left.append(x0)
            y_right.extend([y0, y1, y2])
            y_left.append(y0)
        else:
            x_right.append(x2)
            x_left.append(x3)
            y_right.append(y2)
            y_left.append(y3)
        
        # plot box
        if do_middle_lines and i > 0:
            xx,yy = np.array([x0, x1]), np.array([y0, y1])
            smartcarb.map.add_curve(ax, xx, yy, ls='-', lw=lw,
                                    color=color, label=None, alpha=0.5)

    if len(xa) == 1:
        x_left.append(x3)
        y_left.append(y3)
        
    label = 'Plume polygons' if add_label else None
    xx = np.array(x_right + list(reversed(x_left)))
    yy = np.array(y_right + list(reversed(y_left)))
    smartcarb.map.add_curve(ax, xx, yy, ls='-', lw=lw,
                            color=color, label=label)
            
            

def sort_and_remove_nans(y, c, c_std):
    """
    Sort for y and remove nans.
    """
    sort = np.argsort(y)
    y = y[sort]
    c = c[sort]
    c_std = c_std[sort]

    valids = np.isfinite(y) & np.isfinite(c)

    return y[valids], c[valids], c_std[valids]



def fit_two_gaussian(polygon, share_mu=False):

    def co2_func(x, *a):
        """
        """
        q = a[0]
        sigma = a[1]
        mu = a[2]
        return functions.gauss(x, q, sigma, mu)


    def no2_func(x, *a):
        q = a[3]
        sigma = a[1] # use same sigma as for co2
        if share_mu:
            mu = a[2]
        else:
            mu = a[4]
        return functions.gauss(x, q, sigma, mu)


    # sort and remove nans
    y1, c1, c_std1 = sort_and_remove_nans(polygon.y, polygon.CO2, polygon.CO2_std)
    y2, c2, c_std2 = sort_and_remove_nans(polygon.y, polygon.NO2, polygon.NO2_std)

    # concat
    y = np.concatenate([y1, y2])
    c = np.concatenate([c1, c2])
    c_std = np.concatenate([c_std1, c_std2])
    
    
    def cost_function(x, *a):
        return np.concatenate([co2_func(y1, *a), no2_func(y2, *a)])

    if share_mu:
        p0 = np.array([
            100.0,   # CO2 line density (kg m-1)
             10e3,   # sigma of CO2 and NO2 function (m)
              0.0,   # CO2/NO2 center shift
              0.1,   # NO2 line density (kg m-1)
        ])
        # some bounds for parameter
        bounds= [0, 1e3, -20e3, 0], [1000, 100e3, 20e3, 1000]
    else:
        p0 = np.array([
            100.0,   # CO2 line density (kg m-1)
             10e3,   # sigma of CO2 and NO2 function (m)
             0.0,    # CO2 center shift
              0.1,   # NO2 line density (kg m-1)
              0.0    # NO2 center shift
        ])
        # some bounds for parameter
        bounds= [0, 1e3, -20e3, 0, -20e3], [1000, 100e3, 20e3, 1000, 20e3]

    if y.size < p0.size:
        p = np.full_like(p0, np.nan)
        cov_p = np.full((p0.size, p0.size), np.nan)
        sigma = None

    else:
        if np.all(np.isnan(c_std)):
            sigma = None
        elif np.any(np.isnan(c_std)):
            raise ValueError
        else:
            sigma = c_std

        warnings.simplefilter("error", scipy.optimize.OptimizeWarning)
        try:
            p, cov_p = scipy.optimize.curve_fit(cost_function, y, c, p0, sigma=sigma, bounds=bounds,
                                                absolute_sigma=sigma is not None)

        except (scipy.optimize.OptimizeWarning, RuntimeError):
            p = np.full_like(p0, np.nan)
            cov_p = np.full((p0.size, p0.size), np.nan)
            
    # check if any uncertainty is zero
    
    # sometimes curve_fit finds the negative solution (should not happen anymore with bounds)
    if p[0] < 0 and p[1] < 0 and p[3] < 0:
        p[0] *= -1.0
        p[1] *= -1.0
        p[3] *= -1.0
    
    p_std = np.sqrt(cov_p.diagonal())
    
    if y1.size == 0 or np.isclose(p_std[0], 0.0): # no not enough CO2 observations
        p[0] = np.nan
        p_std[0] = np.nan
        
    if y2.size == 0 or np.isclose(p_std[3], 0.0): # no NO2 observations
        p[3] = np.nan
        p_std[3] = np.nan

    
    if np.any(np.isclose(p_std, 0.0)):
        print('Estimated sigma is equal to zero', p_std)
        p[:] = np.nan
        p_std[:] = np.nan

    # fits
    polygon['CO2_fit'] = co2_func(polygon['y_fit'], *p)
    polygon['NO2_fit'] = no2_func(polygon['y_fit'], *p)

    # CO2 line denisty
    polygon['CO2_line_density'] = p[0]
    polygon['CO2_line_density_std'] = np.nan if sigma is None else p_std[0]
    polygon['CO2_line_density'].attrs['units'] = 'Mt yr-1'

    # NO2 line density
    polygon['NO2_line_density'] = p[3]
    polygon['NO2_line_density_std'] = np.nan if sigma is None else p_std[3]
    polygon['NO2_line_density'].attrs['units'] = 'Mt yr-1'
    
    polygon['CO2_line_density_mu'] = p[2]
    polygon['CO2_line_density_mu_std'] = p_std[2]
    
    if share_mu:
        polygon['NO2_line_density_mu'] = p[2]
        polygon['NO2_line_density_mu_std'] = p_std[2]
    else:
        polygon['NO2_line_density_mu'] = p[4]
        polygon['NO2_line_density_mu_std'] = p_std[4]
    
    polygon['CO2_line_density_sigma'] = p[1]
    polygon['CO2_line_density_sigma_std'] = p_std[1]
    polygon['NO2_line_density_sigma'] = p[1]
    polygon['NO2_line_density_sigma_std'] = p_std[1]
    
    return polygon




def fit_function(y, c, c_std, s=None, function='gauss'):


    # sort
    sort = np.argsort(y)
    y = y[sort]
    c = c[sort]
    c_std = c_std[sort]

    # remove nans
    valids = np.isfinite(c)
    y = y[valids]
    c = c[valids]
    c_std = c_std[valids]

    if function in ['gaussian', 'gauss']:
        curve_function = functions.gauss
        p0 = np.array([
            100.0,  # integral
            10e3,   # sigma
            0.0     # shift
        ])

    elif function == 'erf':
        curve_function = functions.error_function
        p0 = np.array([
            100.0, # integral
            1e3,   # sigma
            20e3,  # box width
            0.0    # shift
        ])
    else:
        raise ValueError

    if np.all(np.isnan(c_std)):
        sigma = None
    else:
        sigma = c_std

    if y.size < p0.size:
        return np.nan, np.nan, np.full(s.shape, np.nan)

    try:
        p, cov_p = scipy.optimize.curve_fit(curve_function, y, c, p0, sigma=sigma,
                                            absolute_sigma=sigma is not None)
    except RuntimeError as e:
        return np.nan, np.nan, np.full(s.shape, np.nan)

    if function in ['gaussian', 'gauss',  'erf']:
        integral = p[0]
        integral_std = np.sqrt(cov_p[0,0])
    else:
        raise ValueError

    return integral, integral_std, curve_function(s, *p)



def standard_error_of_the_mean(x):

    x = np.asarray(x)
    n = np.isfinite(x).sum()

    if n > 0:
        x = np.nanmean(x)
        return x / np.sqrt(n)

    return np.nan


def extract_pixels(data, tracer, xa, xb, ya, yb, dy=None, get_wind=False,
                   verbose=False):
    """
    Extract pixels within a polygon given by plume coords for along-plume
    direction [xa,xb] and across-plume direction [ya,yb] (units: meters).
    """
    name = tracer[:3]
    polygon = xarray.Dataset()

    # only use pixels that are valid observations
    xp = data['xp'].values
    yp = data['yp'].values

    area = (xa <= xp) & (xp < xb) & (ya <= yp) & (yp <= yb)
    
    polygon['polygon_mask'] = xarray.DataArray(area, dims=data.xp.dims)
    polygon['xa'] = xa
    polygon['xb'] = xb
    polygon['ya'] = ya
    polygon['yb'] = yb

    isfinite = data['%s_isfinite' % tracer[:3]].values[area]
    
    if 'other_sources' in data and np.any(data['other_sources'].values[area]):
        if verbose:
            print('Mask other plume pixels in polygon [%f,%f].' % (xa,xb))
        isfinite[data['other_sources'].values[area]] = False
    
    # pixel in area
    polygon['x'] = xarray.DataArray(data.xp.values[area], name='x',
                                    dims='pixels')
    polygon['y'] = xarray.DataArray(data.yp.values[area], name='y',
                                    dims='pixels')

    c = data[tracer].values[area]
    c[~isfinite] = np.nan
    p = data['detected_plume'].values[area]
    
    
    
    polygon[name] = xarray.DataArray(c, name=name, dims='pixels')
    polygon['is_plume'] = xarray.DataArray(p, name='is_plume', dims='pixels')

    if get_wind:
        if 'wind_eff_%s' % tracer[:3] in data:
            wind = data['wind_eff_%s' % tracer[:3]].values[area]
        elif 'u_eff_%s' % tracer[:3] in data:
            ueff = data['u_eff_%s' % tracer[:3]].values[area]
            veff = data['v_eff_%s' % tracer[:3]].values[area]
            wind = np.sqrt(ueff**2 + veff**2)
        else:
            wind = np.full(isfinite.shape, np.nan)

        wind[~isfinite] = np.nan

        polygon['wind'] = xarray.DataArray(wind, name='wind', dims='pixels')



    # estimate noise
    noise_level = data[tracer].attrs.get('noise_level', np.nan)

    if isinstance(noise_level, str):
        molar_mass = 46.0055e-3 # of NO2
        noise_level = {
            'low':  1e15 * molar_mass / N_A * 1e4, # molec cm-2 -> kg m-2
            'high': 2e15 * molar_mass / N_A * 1e4,
            'S5':   1.3e15 * molar_mass / N_A * 1e4
        }[noise_level]

    c_std = np.full(np.shape(c), noise_level)
    polygon['%s_std' % name] = xarray.DataArray(c_std, name='%s_std' % name, dims='pixels')


    polygon['subpolygons'] = xarray.DataArray(
        np.arange(ya + 0.5 * dy, yb, dy), dims='subpolygons'
    )

    sub_areas = [(y0 - 0.5 * dy <= polygon['y']) &
                 (polygon['y'] < y0 + 0.5 * dy)
                 for y0 in polygon['subpolygons']]

    xx = get_values_from_areas(polygon['x'], sub_areas)
    yy = get_values_from_areas(polygon['y'], sub_areas)
    
    c = get_values_from_areas(polygon[name], sub_areas)
    c_std = get_values_from_areas(polygon['%s_std'%name], sub_areas)

    for name in ['x', 'y', name, '%s_std'%name, 'is_plume']:

        if name == 'is_plume':
            function = np.sum
        elif name == '%s_std' % name:
            function = standard_error_of_the_mean
        else:
            function = np.nanmean

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', r'Mean of empty slice')
            values = [function(v) for v in get_values_from_areas(polygon[name].values, sub_areas)]

        polygon['%s_sub' % name] = xarray.DataArray(values, dims='subpolygons')

    return polygon




def get_values_from_areas(values, sub_areas):
    return np.array([values[a.values] if np.any(a) else np.array([]) for a in sub_areas])



def extract_true_line_density(data, tracer, xa, xb, ya, yb):
    """
    """
    # only use pixels that are valid observations
    xp = data['xp'].values
    yp = data['yp'].values

    # use area of true plume if larger than [ya,yb]
    area_true = (xa <= xp) & (xp < xb) & data['%s_true_plume' % tracer[:3]].values

    if np.any(area_true):
        #ya_true, yb_true = get_plume_width(data, dy=5e3, area=area_true)
        ya_true, yb_true = -40e3, 40e3
        ya = min(ya, ya_true)
        yb = max(yb, yb_true)

    # area
    area = (xa <= xp) & (xp < xb) & (ya <= yp) & (yp <= yb)

    # get wind field
    if 'wind_eff_%s' % tracer[:3] in data:
        wind = data['wind_eff_%s' % tracer[:3]].values[area]
    else:
        ueff = data['u_eff_%s' % tracer[:3]].values[area]
        veff = data['v_eff_%s' % tracer[:3]].values[area]
        wind = np.sqrt(ueff**2 + veff**2)

    # get data
    tracer = '%s_true_signal_mass' % tracer[:3]
    c = data[tracer].values[area]

    # include wind speeds at pixel using weights
    c = (c * wind) / np.nanmean(wind)

    line_density = np.sum(c * 2e3**2 / (xb -xa))


    # wind speed
    if 'wind_eff_%s' % tracer[:3] in data:
        wind = data['wind_eff_%s' % tracer[:3]].values[area]
        wind = np.nanmean(wind)

    else:
        ueff = data['u_eff_%s' % tracer[:3]].values[area]
        veff = data['v_eff_%s' % tracer[:3]].values[area]
        wind = np.nanmean(np.sqrt(ueff**2 + veff**2))

    return line_density, yp[area], c, wind


def get_plume_width(data, dy=5e3, area='detected_plume'):

    # distance from center line
    if isinstance(area, str):
        yp = data['yp'].values[data[area]]
    else:
        yp = data['yp'].values[area]

    ymin = np.floor((yp.min() - 2 * dy) / dy) * dy
    ymax = np.ceil((yp.max() + 2 * dy) / dy) * dy

    return ymin, ymax


def extract_line_density(data, tracer, xa, xb, ya=None, yb=None, dy=5e3,
                         method='both', show=False, show_true=True,
                         share_mu=False, extra_width=5e3, add_errors=False,
                         ax=None, verbose=False):
    """
    Extract line densitiy of tracer in data within polygon given by along-plume
    interval [xa, xb] and across-plume interval [ya, yb]. If ya or yb is None
    estimate plume width from detected pixels. The plume center line is
    described by the `curve`.

    The function uses two methods:
        - fitting a Gaussian curve
        - summing mean values in sub-areas in across-plume direction using
          sub-areas with width `dy`
          
    share_mu: fitting two gaussian share mean value
    """
    if ya is None or yb is None:
        ya, yb = get_plume_width(data, dy=extra_width)
    else:
        ya = ya[0]
        yb = yb[0]
        
    if verbose:
        print('Extract for x in [%.1f,%.1f] and y in [%.1f,%1.f]'
              % (xa,xb,ya,yb))
        
        
    # extract pixels in area
    polygon = extract_pixels(data, tracer, xa, xb, ya, yb, dy=dy, get_wind=True,
                             verbose=verbose)
    polygon['y_fit'] = xarray.DataArray(np.linspace(ya, yb, 501), dims='y_fit')
    polygon['CO2_fit'] = xarray.full_like(polygon['y_fit'], np.nan)
    polygon['NO2_fit'] = xarray.full_like(polygon['y_fit'], np.nan)
    polygon['method'] = method
   
    # NO2 values
    if tracer.replace('CO2', 'NO2') in data:
        tracers = ['CO2', 'NO2']
        polygon.update(extract_pixels(data, tracer.replace('CO2', 'NO2'), xa, xb,
                       ya, yb, dy=dy, get_wind=True))
    else:
        tracers = ['CO2']
    

    if method in ['gauss', 'gaussian']:

        for tracer in tracers:

            # fit function (Gaussian or Error function)
            if np.any((polygon['is_plume_sub'] > 0) & np.isnan(polygon['%s_sub' % tracer])):
                polygon['%s_line_density' % tracer] = np.nan
                polygon['%s_line_density' % tracer] = np.nan
            else:
                mass_gauss, mass_gauss_std, fit_gauss = fit_function(polygon['y'],
                                                                     polygon[tracer],
                                                                     polygon['%s_std' % tracer],
                                                                     polygon['y_fit'],
                                                                     function='gauss')

                polygon['%s_line_density' % tracer] = mass_gauss
                polygon['%s_line_density_std' % tracer] = mass_gauss_std
                polygon['%s_fit' % tracer][:] = fit_gauss

            # set not used parameters to nan
            polygon['%s_line_density_mu' % tracer] = np.nan
            polygon['%s_line_density_sigma' % tracer] = np.nan
            polygon['%s_line_density_mu_std' % tracer] = np.nan
            polygon['%s_line_density_sigma_std' % tracer] = np.nan
                

    elif method in ['sub-areas']:

        for tracer in tracers:

            if np.any((polygon['is_plume_sub'] > 0) & np.isnan(polygon['%s_sub' % tracer])):
                polygon['%s_line_density' % tracer] = np.nan
                polygon['%s_line_density' % tracer] = np.nan
            else:

                valids = np.isfinite(polygon['CO2_sub'].values)

                ss = polygon['subpolygons'].values
                area_means = polygon['%s_sub' % tracer].values
                area_means_std = polygon['%s_std_sub' % tracer].values

                if np.all(~valids):
                    area_means[:] = np.nan
                else:
                    area_means = np.interp(ss, ss[valids], area_means[valids],
                                           left=0.0, right=0.0)

                polygon['%s_line_density' % tracer] = np.sum(area_means * dy)
                
                # FIXME
                n = np.sum(valids)
                polygon['%s_line_density_std' % tracer] = np.sqrt( np.nansum(area_means_std**2 * dy**2) ) / np.sqrt(n)

            # set not used parameters to nan
            polygon['%s_line_density_mu' % tracer] = np.nan
            polygon['%s_line_density_sigma' % tracer] = np.nan
            polygon['%s_line_density_mu_std' % tracer] = np.nan
            polygon['%s_line_density_sigma_std' % tracer] = np.nan

    elif method in ['two-gauss']:
        polygon = fit_two_gaussian(polygon, share_mu=share_mu)
        
    elif method == 'none':
        pass
        
    else:
        raise ValueError

    # x_sub and y_sub have only nans if no pixels are inside subpolygon, 
    # i.e. sub-polygon is not in swath
    # FIXME: does not work when plume is wider than plume area (~10 km)
    #not_fully_in_swath = np.any(np.isnan(polygon['x_sub']) & (np.abs(polygon['subpolygons']) <= 10.0))
    not_fully_in_swath = np.any(np.isnan(polygon['x_sub']))
    
    #if verbose and not_fully_in_swath:
    if not_fully_in_swath:
        print('Polygon not fully in swath')
        
    # only use fits with at least one observation per sub-polygon, 
    # if pixels have been detected
    for tracer in tracers:
        if not_fully_in_swath or (np.any((polygon['is_plume_sub'] > 0)
                                         & np.isnan(polygon['%s_sub' % tracer]))):
            
            if verbose:
                print('Set %s LD to nan: not in swath or no pixels in sub-polygon with plume detection' % tracer)
            
            polygon['%s_line_density' % tracer] = np.nan
            polygon['%s_line_density_std' % tracer] = np.nan
            
            if method == 'two-gauss':
                polygon['%s_fit' % tracer][:] = np.nan


    # set unrealistic low/high values to nan
    if method != 'none':
        if (
            polygon[f'{tracer}_line_density'] < -1000.0 or polygon[f'{tracer}_line_density'] > 1000.0 or
            polygon[f'{tracer}_line_density_std'] > 1000.0
        ):
            if verbose:
                print('Fit failed:', polygon[f'{tracer}_line_density'], polygon[f'{tracer}_line_density_std'])
            polygon[f'{tracer}_line_density'] = np.nan
            polygon[f'{tracer}_line_density_std'] = np.nan
            if method == 'two-gauss':
                polygon[f'{tracer}_fit'][:] = np.nan


    if show:
        emiss.vis.plot_across_section(polygon, tracer=tracer[:3],
                                      method=method,
                                      show_true=show_true,
                                      add_errors=add_errors, ax=ax)

    return polygon







